@@ -10,10 +10,8 @@ from itertools import groupby
 from sql import Literal, Union, Column, Null, For
 from sql.aggregate import Sum
 from sql.conditionals import Coalesce, Case
-from sql.operators import Concat
 
 from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, Check
-from trytond import backend
 from trytond.pyson import Eval, If, Bool
 from trytond.tools import reduce_ids
 from trytond.transaction import Transaction
@@ -59,18 +57,18 @@ class StockMixin(object):
         return new_context
 
     @classmethod
-    def _get_quantity(cls, records, name, location_ids, products=None,
-            grouping=('product',), position=-1):
+    def _get_quantity(cls, records, name, location_ids,
+            grouping=('product',), grouping_filter=None, position=-1):
         """
         Compute for each record the stock quantity in the default uom of the
         product.
 
         location_ids is the list of IDs of locations to take account to compute
             the stock.
-        products restrict the stock computation to the this products (more
-            efficient), so it should be the products related to records.
-            If it is None all products are used.
         grouping defines how stock moves are grouped.
+        grouping_filter is a tuple of values, for the Move's field at the same
+            position in grouping tuple, used to filter which moves are used to
+            compute quantities. If it is None all the products are used.
         position defines which field of grouping corresponds to the record
             whose quantity is computed.
 
@@ -84,16 +82,17 @@ class StockMixin(object):
         if not location_ids:
             return quantities
 
-        product_ids = products and [p.id for p in products] or None
         with_childs = Transaction().context.get(
             'with_childs', len(location_ids) == 1)
 
         with Transaction().set_context(cls._quantity_context(name)):
-            pbl = Product.products_by_location(location_ids=location_ids,
-                product_ids=product_ids, with_childs=with_childs,
-                grouping=grouping)
+            pbl = Product.products_by_location(
+                location_ids,
+                with_childs=with_childs,
+                grouping=grouping,
+                grouping_filter=grouping_filter)
 
-        for key, quantity in pbl.iteritems():
+        for key, quantity in pbl.items():
             # pbl could return None in some keys
             if (key[position] is not None and
                     key[position] in quantities):
@@ -127,8 +126,7 @@ class StockMixin(object):
 
         with Transaction().set_context(cls._quantity_context(name)):
             pbl = Product.products_by_location(
-                location_ids=location_ids, with_childs=with_childs,
-                grouping=grouping)
+                location_ids, with_childs=with_childs, grouping=grouping)
 
         _, operator_, operand = domain
         operator_ = {
@@ -142,7 +140,7 @@ class StockMixin(object):
             'not in': lambda v, l: v not in l,
             }.get(operator_, lambda v, l: False)
         record_ids = []
-        for key, quantity in pbl.iteritems():
+        for key, quantity in pbl.items():
             if operator_(quantity, operand):
                 # pbl could return None in some keys
                 if key[position] is not None:
@@ -294,88 +292,42 @@ class Move(Workflow, ModelSQL, ModelView):
                 'cancel': {
                     'invisible': ~Eval('state').in_(['draft', 'assigned']),
                     'readonly': Eval('shipment'),
+                    'depends': ['state', 'shipment'],
                     },
                 'draft': {
                     'invisible': ~Eval('state').in_(['assigned']),
                     'readonly': Eval('shipment'),
+                    'depends': ['state', 'shipment'],
                     },
                 'assign': {
                     'invisible': ~Eval('state').in_(['assigned']),
+                    'depends': ['state'],
                     },
                 'do': {
                     'invisible': ~Eval('state').in_(['draft', 'assigned']),
                     'readonly': (Eval('shipment')
                         | (Eval('assignation_required', True)
                             & (Eval('state') == 'draft'))),
+                    'depends': ['state', 'assignation_required', 'shipment'],
                     },
                 })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        transaction = Transaction()
-        cursor = transaction.connection.cursor()
         sql_table = cls.__table__()
 
-        # Migration from 1.2: packing renamed into shipment
-        table = TableHandler(cls, module_name)
-        table.drop_constraint('check_packing')
-        for suffix in ('in', 'out', 'in_return', 'out_return', 'internal'):
-            old_column = 'packing_%s' % suffix
-            new_column = 'shipment_%s' % suffix
-            if table.column_exist(old_column):
-                table.index_action(old_column, action='remove')
-            table.drop_fk(old_column)
-            table.column_rename(old_column, new_column)
-
-        # Migration from 1.8: new field internal_quantity
-        internal_quantity_exist = table.column_exist('internal_quantity')
-
         super(Move, cls).__register__(module_name)
-
-        # Migration from 1.8: fill new field internal_quantity
-        if not internal_quantity_exist:
-            offset = 0
-            limit = transaction.database.IN_MAX
-            moves = True
-            while moves:
-                moves = cls.search([], offset=offset, limit=limit)
-                offset += limit
-                for move in moves:
-                    internal_quantity = cls._get_internal_quantity(
-                            move.quantity, move.uom, move.product)
-                    cursor.execute(*sql_table.update(
-                            columns=[sql_table.internal_quantity],
-                            values=[internal_quantity],
-                            where=sql_table.id == move.id))
-            table = TableHandler(cls, module_name)
-            table.not_null_action('internal_quantity', action='add')
-
-        # Migration from 1.0 check_packing_in_out has been removed
-        table = TableHandler(cls, module_name)
-        table.drop_constraint('check_packing_in_out')
-
-        # Migration from 2.6: merge all shipments
-        table.drop_constraint('check_shipment')
-        shipments = {
-            'shipment_in': 'stock.shipment.in',
-            'shipment_out': 'stock.shipment.out',
-            'shipment_out_return': 'stock.shipment.out.return',
-            'shipment_in_return': 'stock.shipment.in.return',
-            'shipment_internal': 'stock.shipment.internal',
-            }
-        for column, model in shipments.iteritems():
-            if table.column_exist(column):
-                cursor.execute(*sql_table.update(
-                        columns=[sql_table.shipment],
-                        values=[Concat(model + ',',
-                                Column(sql_table, column))],
-                        where=Column(sql_table, column) != Null))
-                table.drop_column(column)
+        table = cls.__table_handler__(module_name)
 
         # Add index on create_date
         table.index_action('create_date', action='add')
 
+        # Index for period join in compute_quantities_query
+        table.index_action([
+                Coalesce(sql_table.effective_date,
+                    sql_table.planned_date,
+                    datetime.date.max)], action='add')
+
     @staticmethod
     def default_planned_date():
         return Transaction().context.get('planned_date')
@@ -527,7 +479,7 @@ class Move(Workflow, ModelSQL, ModelView):
                 for move in moves:
                     date = (move.effective_date if move.effective_date
                         else move.planned_date)
-                    if date and date < period.date:
+                    if date and date <= period.date:
                         cls.raise_user_error('period_closed', {
                                 'move': move.rec_name,
                                 'period': period.rec_name,
@@ -786,7 +738,7 @@ class Move(Workflow, ModelSQL, ModelView):
             return ((move.from_location.type in types) ^
                 (move.to_location.type in types)
                 and not move.origin)
-        moves = filter(no_origin, moves)
+        moves = list(filter(no_origin, moves))
         if moves:
             names = ', '.join(m.rec_name for m in moves[:5])
             if len(moves) > 5:
@@ -803,7 +755,7 @@ class Move(Workflow, ModelSQL, ModelView):
         """
         to_pick = []
         needed_qty = self.quantity
-        for location, available_qty in location_quantities.iteritems():
+        for location, available_qty in location_quantities.items():
             # Ignore available_qty when too small
             if available_qty < self.uom.rounding:
                 continue
@@ -880,9 +832,9 @@ class Move(Workflow, ModelSQL, ModelView):
                 stock_date_end=stock_date_end,
                 stock_assign=True):
             pbl = Product.products_by_location(
-                location_ids=location_ids,
-                product_ids=product_ids,
-                grouping=grouping)
+                location_ids,
+                grouping=grouping,
+                grouping_filter=(product_ids,))
 
         def get_key(move, location):
             key = (location.id,)
@@ -946,7 +898,8 @@ class Move(Workflow, ModelSQL, ModelView):
                     to_assign.append(move)
                     first = False
                 else:
-                    to_assign.extend(cls.copy([move], default=values))
+                    with Transaction().set_context(_stock_move_split=True):
+                        to_assign.extend(cls.copy([move], default=values))
 
                 qty_default_uom = Uom.compute_qty(move.uom, qty,
                         move.product.default_uom, round=False)
@@ -980,8 +933,8 @@ class Move(Workflow, ModelSQL, ModelView):
                 quantities of all child locations but quantities of the storage
                 zone.
         If with_childs, it computes also for child locations.
-        grouping is a tuple of Move field names and defines how stock moves are
-            grouped.
+        grouping is a tuple of Move (or Product if prefixed by 'product.')
+            field names and defines how stock moves are grouped.
         grouping_filter is a tuple of values, for the Move's field at the same
             position in grouping tuple, used to filter which moves are used to
             compute quantities. It must be None or have the same number of
@@ -997,6 +950,7 @@ class Move(Workflow, ModelSQL, ModelView):
         Date = pool.get('ir.date')
         Period = pool.get('stock.period')
         Move = pool.get('stock.move')
+        Product = pool.get('product.product')
 
         move = Move.__table__()
         today = Date.today()
@@ -1005,17 +959,52 @@ class Move(Workflow, ModelSQL, ModelView):
             return None
         context = Transaction().context.copy()
 
+        use_product = False
         for field in grouping:
-            if field not in Move._fields:
-                raise ValueError('"%s" has no field "%s"' % (Move, field))
-        assert grouping_filter is None or len(grouping_filter) == len(grouping)
+            if field.startswith('product.'):
+                Model = Product
+                field = field[len('product.'):]
+                use_product = True
+            else:
+                Model = Move
+            if field not in Model._fields:
+                raise ValueError('"%s" has no field "%s"' % (Model, field))
+        assert grouping_filter is None or len(grouping_filter) <= len(grouping)
+        assert len(set(grouping)) == len(grouping)
 
         move_rule_query = Rule.query_get('stock.move')
 
+        def get_column(name, table, product):
+            if name.startswith('product.'):
+                column = Column(product, name[len('product.'):])
+            else:
+                column = Column(table, name)
+            return column.as_(name)
+
+        if use_product:
+            product = Product.__table__()
+            columns = ['id', 'state', 'effective_date', 'planned_date',
+            'internal_quantity', 'from_location', 'to_location']
+            columns += list(grouping)
+            columns = [get_column(c, move, product) for c in columns]
+            move = (move
+                .join(product, condition=move.product == product.id)
+                .select(*columns))
+
         PeriodCache = Period.get_cache(grouping)
         period = None
         if PeriodCache:
             period_cache = PeriodCache.__table__()
+            if use_product:
+                product_cache = Product.__table__()
+                columns = ['internal_quantity', 'period', 'location']
+                columns += list(grouping)
+                columns = [get_column(c, period_cache, product_cache)
+                    for c in columns]
+                period_cache = (period_cache
+                    .join(product_cache,
+                        condition=period_cache.product == product_cache.id)
+                    .select(*columns))
 
         if with_childs:
             # Replace tables with union which replaces flat children locations
@@ -1223,7 +1212,7 @@ class Move(Workflow, ModelSQL, ModelView):
                     context.get('stock_assign'))
         elif PeriodCache:
             periods = Period.search([
-                    ('date', '<', context['stock_date_end']),
+                    ('date', '<=', context['stock_date_end']),
                     ('state', '=', 'closed'),
                     ], order=[('date', 'DESC')], limit=1)
             if periods:
@@ -1249,7 +1238,7 @@ class Move(Workflow, ModelSQL, ModelView):
                 column = Column(move, fieldname)
                 if PeriodCache:
                     cache_column = Column(period_cache, fieldname)
-                if isinstance(grouping_ids[0], (int, long, float, Decimal)):
+                if isinstance(grouping_ids[0], (int, float, Decimal)):
                     where &= reduce_ids(column, grouping_ids)
                     if PeriodCache:
                         where_period &= reduce_ids(cache_column, grouping_ids)
@@ -1333,17 +1322,24 @@ class Move(Workflow, ModelSQL, ModelView):
         """
         pool = Pool()
         Product = pool.get('product.product')
+        Template = pool.get('product.template')
 
         assert query is not None, (
             "Query in Move.compute_quantities() can't be None")
-        assert 'product' in grouping
+        assert 'product' in grouping or 'product.template' in grouping
 
         cursor = Transaction().connection.cursor()
         cursor.execute(*query)
         raw_lines = cursor.fetchall()
 
-        product_getter = operator.itemgetter(grouping.index('product') + 1)
-        res_product_ids = set()
+        if 'product' in grouping:
+            id_name = 'product'
+            Model = Product
+        else:
+            id_name = 'product.template'
+            Model = Template
+        id_getter = operator.itemgetter(grouping.index(id_name) + 1)
+        ids = set()
         quantities = defaultdict(lambda: 0)
         keys = set()
         # We can do a quick loop without propagation if the request is for a
@@ -1357,7 +1353,7 @@ class Move(Workflow, ModelSQL, ModelView):
             key = tuple(line[1:-1])
             quantity = line[-1]
             quantities[(location,) + key] += quantity
-            res_product_ids.add(product_getter(line))
+            ids.add(id_getter(line))
             keys.add(key)
 
         # Propagate quantities on from child locations to their parents
@@ -1393,18 +1389,17 @@ class Move(Workflow, ModelSQL, ModelView):
                 leafs = next_leafs
 
             # clean result
-            for key in quantities.keys():
+            for key in list(quantities.keys()):
                 location = key[0]
                 if location not in location_ids:
                     del quantities[key]
 
         # Round quantities
         default_uom = dict((p.id, p.default_uom) for p in
-            Product.browse(list(res_product_ids)))
-        for key, quantity in quantities.iteritems():
+            Model.browse(list(ids)))
+        for key, quantity in quantities.items():
             location = key[0]
-            product = product_getter(key)
-            uom = default_uom[product]
+            uom = default_uom[id_getter(key)]
             quantities[key] = uom.round(quantity)
 
         return quantities
